from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.agent import Agent
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_AGENT(state: State, name: str) -> None:    
    state.require(not state.in_agent, 'Already inside an agent.', 'First end current agent using EAGENT.')
    state.require(
        not state.in_message, 
        'Cannot define agents inside messages.', 
        'First end current message using EMESSAGE.'
    )
    state.require(not state.in_graph, 'Cannot define agents inside graphs.', 'First end current graph using EGRAPH.')
    state.require(not state.agent_exists(name), f'Agent {name} already exists in the current environment.')
    state.require(
        is_valid_name(name), 
        f'{name} is not a correct name.', 
        f'Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.'
    )
    
    state.in_agent = True
    state.add_agent(Agent(name))


def op_EAGENT(state: State) -> None:    
    state.require(state.in_agent, 'Not inside any agent.', 'Try defining new agents using AGENT.')
    state.require(
        not state.in_behaviour, 
        'Cannot end an agent inside a behaviour.', 
        'First end current behaviour using EBEHAV.'
    )
    
    state.in_agent = False



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from aasm.intermediate.action import Action
    from aasm.intermediate.message import Message


class Behaviour:
    def __init__(self, name: str):
        self.name = name
        self.actions: Dict[str, Action] = {}
        
    @property
    def last_action(self) -> Action:
        return self.actions[list(self.actions.keys())[-1]]
    
    def add_action(self, action: Action) -> None:
        self.actions[action.name] = action
        
    def action_exists(self, name: str) -> bool:
        return name in self.actions

    def print(self) -> None:
        print(f'Behaviour {self.name}')
        for action in self.actions.values():
            action.print()


class SetupBehaviour(Behaviour):
    def __init__(self, name: str):
        super().__init__(name)
        
        
class OneTimeBehaviour(Behaviour):
    def __init__(self, name: str, delay: str):
        super().__init__(name)
        self.delay: str = delay
        
        
class CyclicBehaviour(Behaviour):
    def __init__(self, name: str, period: str):
        super().__init__(name)
        self.period: str = period
        
        
class MessageReceivedBehaviour(Behaviour):
    def __init__(self, name: str, message: Message):
        super().__init__(name)
        self.received_message: Message = message



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import (IfEqual, IfGreaterThan,
                                           IfGreaterThanOrEqual, IfLessThan,
                                           IfLessThanOrEqual, IfNotEqual,
                                           WhileEqual, WhileGreaterThan,
                                           WhileGreaterThanOrEqual,
                                           WhileLessThan, WhileLessThanOrEqual,
                                           WhileNotEqual)

if TYPE_CHECKING:
    from aasm.parsing.state import State


def handle_unordered_conditional_statement(state: State, op: str, arg1: str, arg2: str) -> None:
    state.require(state.in_action, 'Not inside any action.', f'{op} can be used inside actions.')     
    lhs = Argument(state, arg1)
    rhs = Argument(state, arg2)
    state.require(
        lhs.unordered_comparaison_context(rhs), 
        'Mismatched types in the unordered comparaison context.', 
        f'ARG1 {lhs.explain()}, ARG2 {rhs.explain()}'
    )
    
    match op:
        case 'IEQ':
            state.last_action.add_instruction(IfEqual(lhs, rhs))

        case 'INEQ':
            state.last_action.add_instruction(IfNotEqual(lhs, rhs))

        case 'WEQ':
            state.last_action.add_instruction(WhileEqual(lhs, rhs))

        case 'WNEQ':
            state.last_action.add_instruction(WhileNotEqual(lhs, rhs))

        case _:
            state.panic(f'Unexpected error: {op} {arg1} {arg2}')
    
    state.last_action.start_block()


def handle_ordered_conditional_statement(state: State, op: str, arg1: str, arg2: str) -> None:
    state.require(state.in_action, 'Not inside any action.', f'{op} can be used inside actions.')     
    lhs = Argument(state, arg1)
    rhs = Argument(state, arg2)
    state.require(
        lhs.ordered_comparaison_context(rhs),
        'Mismatched types in the ordered comparaison context.', 
        f'ARG1 {lhs.explain()}, ARG2 {rhs.explain()}'
    )
    
    match op:
        case 'IGT':
            state.last_action.add_instruction(IfGreaterThan(lhs, rhs))

        case 'IGTEQ':
            state.last_action.add_instruction(IfGreaterThanOrEqual(lhs, rhs))

        case 'ILT':
            state.last_action.add_instruction(IfLessThan(lhs, rhs))

        case 'ILTEQ':
            state.last_action.add_instruction(IfLessThanOrEqual(lhs, rhs))

        case 'WGT':
            state.last_action.add_instruction(WhileGreaterThan(lhs, rhs))

        case 'WGTEQ':
            state.last_action.add_instruction(WhileGreaterThanOrEqual(lhs, rhs))

        case 'WLT':
            state.last_action.add_instruction(WhileLessThan(lhs, rhs))

        case 'WLTEQ':
            state.last_action.add_instruction(WhileLessThanOrEqual(lhs, rhs))

        case _:
            state.panic(f'Unexpected error: {op} {arg1} {arg2}')
    
    state.last_action.start_block()
